# Yuehao Bai | University of Michigan | Jan 2021
# ----------------------------------------------
# Functions for inference for experiments with matched pairs
# -----------------------------------------------
library(MASS)
library(compiler)
library(nbpMatching)

# We use the R compiler to improve the speed of the functions.
# ----------------------------------------------
# List of functions:
# ----------------------------------------------

# A function to convert factor to numeric.
# ---------------------------------------------------------
as.double.factor <- function(x) {
  as.double(as.character(x))
}
# ---------------------------------------------------------

genBaseline <-
  function(n,
           model,
           theta = 0,
           gamma,
           sg1 = 1,
           rho = 0.2,
           d = 5,
           error = "N") {
    # ----------------------------------------------
    # Generate covariates and potential outcomes using different models.
    # Models indexed by 1--15. To get models 13--15, just set models to 10--12 and d to 100.
    # gamma: scalar for models 1--6, bivariate for models 7--9.
    # rho: only needed for models 7--9.
    # d: only needed for models 10--15. d = 5 for models 10--12, d = 100 for models 13--15.
    # error: dist. of errors, "N" or "t" (df = 4).
    # Normalized so that E[m_1(X_i)] = E[m_0(X_i)] = 0 and theta_0 = 0.
    # ----------------------------------------------
    mu.1 <- theta  # normalize mu.0 = 0
    # data is generated as Y.0 = 0 + m.0 + e.0, Y.1 = mu.1 + m.1 + e.1
    
    if (model <= 6) {
      X <- runif(2 * n)
      # X <- rbeta(2 * n, 1, 1)
    }

    if (error == "N") {
      e.0.raw <- rnorm(2 * n)
      e.1.raw <- sqrt(sg1) * rnorm(2 * n)
    }

    if (error == "t") {
      e.0.raw <- rt(2 * n, df = 4)
      e.1.raw <- sqrt(sg1) * rt(2 * n, df = 4)
    }
    
    if (model == 1) {
      # homogeneous, linear model and uniform X
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- gamma * (X - 1 / 2)  # make sure E[m0] = 0
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m0 + e.1
    }
    
    if (model == 2) {
      # homogeneous, sine model
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- sin(gamma * (X - 1 / 2))
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m0 + e.1
    }
    
    if (model == 3) {
      # heterogeneous, sine model
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- sin(gamma * (X - 1 / 2))
      m1 <- m0 + X ^ 2 - 1 / 3
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 4) {
      # a more pronounced heterogeneous model
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- 0
      m1 <- 0 + 10 * (X ^ 2 - 1 / 3)
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 5) {
      # heterogeneous model with E[Y_i(1)+Y_i(0)|X_i] = E[Y_i(1)+Y_i(0)]
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- 10 * (X ^ 2 - 1 / 3)
      m1 <- -m0
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 6) {
      # heterogeneous, heteroskedastic model
      e.0 <- X ^ 2 * e.0.raw
      e.1 <- X ^ 2 * e.1.raw
      m0 <- 0
      m1 <- 0 + 10 * (X ^ 2 - 1 / 3)
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 7) {
      # bivariate, homogeneous, linear model
      X <- mvrnorm(2 * n, c(0, 0), matrix(c(1, rho, rho, 1), 2, 2))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <-
        gamma[1] * (X[, 1] - 1 / 2) + gamma[2] * (X[, 2] - 1 / 2)
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m0 + e.1
    }
    
    if (model == 8) {
      # bivariate, heterogeneous model
      X <- mvrnorm(2 * n, c(0, 0), matrix(c(1, rho, rho, 1), 2, 2))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <-
        gamma[1] * (X[, 1] - 1 / 2) + gamma[2] * (X[, 2] - 1 / 2)
      m1 <- m0 + 10 * (qnorm(X[, 1]) * qnorm(X[, 2]) - rho)
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 9) {
      # bivariate, heterogeneous model with E[Y_i(1)+Y_i(0)|X_i] = E[Y_i(1)+Y_i(0)]
      X <- mvrnorm(2 * n, c(0, 0), matrix(c(1, rho, rho, 1), 2, 2))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- 5 * (qnorm(X[, 1]) * qnorm(X[, 2]) - rho)
      m1 <- -m0
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model == 10) {
      X <- mvrnorm(2 * n, rep(0, d), matrix(rho, d, d) + (1 - rho) * diag(d))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- (X - 1 / 2) %*% gamma
      m1 <- m0
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }

    if (model == 11) {
      X <- mvrnorm(2 * n, rep(0, d), matrix(rho, d, d) + (1 - rho) * diag(d))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- (X - 1 / 2) %*% gamma
      m1 <- m0 + 10 * (qnorm(X) %*% rep(1, d))
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }

    if (model == 12) {
      X <- mvrnorm(2 * n, rep(0, d), matrix(rho, d, d) + (1 - rho) * diag(d))
      X <- pnorm(X)
      e.0 <- e.0.raw
      e.1 <- e.1.raw
      m0 <- 5 * (qnorm(X) %*% rep(1, d))
      m1 <- - m0
      Y.0 <- m0 + e.0
      Y.1 <- mu.1 + m1 + e.1
    }
    
    if (model <= 6) {
      X <- as.matrix(X)
    }
    
    list(
      n = n,
      model = model,
      Y.0 = Y.0,
      Y.1 = Y.1,
      X = X,
      v.0 = m0,
      v.1 = mu.1 + m1
    )
  }

matchNbp <- function(X) {
  n <- dim(X)[1] / 2 ## Remember / 2! Previously forgot / 2 but results same.
  if (dim(X)[2] == 1) {
    pairs.vec <- order(X)
    pairs <- matrix(pairs.vec, n, 2, byrow = T)
  } else {
    dist.mat <- as.matrix(dist(X, diag = T, upper = T))
    dist <- distancematrix(dist.mat)
    match <- nonbimatch(dist)
    pairs <- match$halves
    pairs <-
      cbind(as.double.factor(pairs[, 1]), as.double.factor(pairs[, 3]))
  }
}

genData <- function(baseline, method = 1) {
  # ---------------------------------------
  # Generate the observations. Notice that data are reindexed according to pairs.
  # ---------------------------------------
  n <- baseline$n
  X <- baseline$X
  Y.0 <- baseline$Y.0
  Y.1 <- baseline$Y.1
  pairs <- matchNbp(baseline$X)
  pairs.vec <-
    matrix(t(pairs), 2 * n, 1)  # remember to transpose, and the same for below
  X.pair <- as.matrix(X[pairs.vec,])
  Y.0.pair <- Y.0[pairs.vec]
  Y.1.pair <- Y.1[pairs.vec]
  
  A.mat <- cbind(rep(-1, n), rep(1, n))
  coin.flip <- sample(c(-1, 1), n, replace = T)
  A.mat <- A.mat * coin.flip
  A.mat <- (A.mat + 1) / 2
  A.pair <- matrix(t(A.mat), 2 * n, 1)
  Y.pair <- Y.0.pair * (1 - A.pair) + Y.1.pair * A.pair
  list(n = n,
       X = X.pair,
       A = A.pair,
       Y = Y.pair)
}

genDataReg <- function(baseline, method = 1) {
  # ---------------------------------------
  # Generate the observations. Notice that data are reindexed according to pairs.
  # ---------------------------------------
  n <- baseline$n
  X <- baseline$X
  Y.0 <- baseline$Y.0
  Y.1 <- baseline$Y.1
  pairs <- matchNbp(as.matrix(baseline$X[, 1]))
  pairs.vec <-
    matrix(t(pairs), 2 * n, 1)  # remember to transpose, and the same for below
  X.pair <- as.matrix(X[pairs.vec,])
  Y.0.pair <- Y.0[pairs.vec]
  Y.1.pair <- Y.1[pairs.vec]
  
  A.mat <- cbind(rep(-1, n), rep(1, n))
  coin.flip <- sample(c(-1, 1), n, replace = T)
  A.mat <- A.mat * coin.flip
  A.mat <- (A.mat + 1) / 2
  A.pair <- matrix(t(A.mat), 2 * n, 1)
  Y.pair <- Y.0.pair * (1 - A.pair) + Y.1.pair * A.pair
  list(n = n,
       X = X.pair,
       A = A.pair,
       Y = Y.pair)
}

testT <- function(data.obs, alpha.level = 0.05) {
  # -------------------------
  # t-test.
  # -------------------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.c <- Y[A == 0]
  Y.t <- Y[A == 1]
  mean.c <- mean(Y.c)
  mean.t <- mean(Y.t)
  test.num <- sqrt(n) * (mean.t - mean.c)
  var.c <- mean((Y.c - mean.c) ^ 2)
  var.t <- mean((Y.t - mean.t) ^ 2)
  test.den <- sqrt(var.c + var.t)
  test.stat <- test.num / test.den
  test <- abs(test.stat) > qnorm(1 - alpha.level / 2)
  list(test = test, test.num = test.num / sqrt(n), test.den = test.den / sqrt(n), n = n)
}
testT <- cmpfun(testT)

testRand <- function(data.obs, alpha.level = 0.05, n.perm) {
  # ------------------------------------------------------
  # Naive randomization test which flips treatment and control.
  # ------------------------------------------------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.diff <- Y[A == 1] - Y[A == 0]
  perm.stat <- rep(NA, n.perm)
  
  for (i in 1:n.perm) {
    coin.flip <- sample(c(-1, 1), n, replace = T)
    perm.stat[i] <- abs(mean(Y.diff * coin.flip))
  }
  
  test.stat <- abs(mean(Y.diff))
  perm.stat[n.perm] <- test.stat
  test.cv <- quantile(perm.stat, 1 - alpha.level)
  test <- test.stat > test.cv
}
testRand <- cmpfun(testRand)

testSC <- function(data.obs, alpha.level = 0.05) {
  # --------------------------------------------------
  # Test with correct studentization. In 1 dim, need to order the data correctly.
  # --------------------------------------------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.diff <- Y[A == 1] - Y[A == 0]
  test.num <- sqrt(n) * mean(Y.diff)
  tau2.hat <- mean(Y.diff ^ 2)
  
  if (dim(X)[2] == 1) {
    phi2.hat <-
      mean(Y.diff[seq(1, n - 1, by = 2)] * Y.diff[seq(2, n, by = 2)])
  } else {
    X.mid <- (X[seq(1, n - 1, by = 2), ] + X[seq(2, n, by = 2), ]) / 2
    mid.pairs <- matchNbp(X.mid)
    phi2.hat <-
      mean(Y.diff[mid.pairs[, 1]] * Y.diff[mid.pairs[, 2]])
  }
  
  var.sc <- tau2.hat - phi2.hat / 2 - (mean(Y.diff))^2 / 2
  test.den <- sqrt(var.sc)
  test.stat <- test.num / test.den
  test <- abs(test.stat) > qnorm(1 - alpha.level / 2)
  list(test = test, test.num = test.num / sqrt(n), test.den = test.den / sqrt(n), n = n)
}
testSC <- cmpfun(testSC)

testSCRand <- function(data.obs, alpha.level = 0.05, n.perm) {
  # ---------------
  # Randomization test with correct studentization.
  # ---------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.diff <- Y[A == 1] - Y[A == 0]
  test.num <- sqrt(n) * mean(Y.diff)
  tau2.hat <- mean(Y.diff ^ 2)
  
  if (dim(X)[2] == 1) {
    Y.diff.1 <- Y.diff[seq(1, n - 1, by = 2)]
    Y.diff.2 <- Y.diff[seq(2, n, by = 2)]
  } else {
    X.mid <- (X[seq(1, n - 1, by = 2), ] + X[seq(2, n, by = 2), ]) / 2
    mid.pairs <- matchNbp(X.mid)
    Y.diff.1 <- Y.diff[mid.pairs[, 1]]
    Y.diff.2 <- Y.diff[mid.pairs[, 2]]
  }

  phi2.hat <- mean(Y.diff.1 * Y.diff.2)  
  var.sc <- tau2.hat - phi2.hat / 2 - (mean(Y.diff)^2) / 2
  test.den <- sqrt(var.sc)
  test.stat <- abs(test.num / test.den)
  perm.stat <- rep(NA, n.perm)
  
  for (i in 1:n.perm) {
    coin.flip <- sample(c(-1, 1), n, replace = T)
    test.num <- sqrt(n) * mean(Y.diff * coin.flip)
    phi2.hat <- mean(Y.diff.1 * Y.diff.2 * coin.flip[seq(1, n - 1, by = 2)] * coin.flip[seq(2, n, by = 2)])
    var.sc <- tau2.hat - phi2.hat / 2 - (test.num / sqrt(n))^2 / 2
    test.den <- sqrt(var.sc)
    perm.stat[i] <- abs(test.num / test.den)
  }

  perm.stat[n.perm] <- test.stat
  test.cv <- quantile(perm.stat, 1 - alpha.level)
  test <- test.stat > test.cv
}
testSCRand <- cmpfun(testSCRand)

testMPt <- function(data.obs, alpha.level = 0.05) {
  # -------------
  # Matched-pair t-test.
  # -------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.diff <- Y[A == 1] - Y[A == 0]
  delta.hat <- mean(Y.diff)
  test.num <- sqrt(n) * delta.hat
  test.den <- sqrt(mean(Y.diff ^ 2) - delta.hat^2)
  test.stat <- test.num / test.den
  test <- abs(test.stat) > qnorm(1 - alpha.level / 2)
  list(test = test, test.num = test.num / sqrt(n), test.den = test.den / sqrt(n), n = n)
}

testSCNew <- function(data.obs, alpha.level = 0.05) {
  # --------------------------------------------------
  # Test with correct studentization, new version.
  # --------------------------------------------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  Y.sum <- Y[A == 1] + Y[A == 0]
  test.num <- sqrt(n) * mean(Y[A == 1] - Y[A == 0])

  if (dim(X)[2] == 1) {
    phi2.hat <-
      mean(Y.sum[seq(1, n - 1, by = 2)] * Y.sum[seq(2, n, by = 2)])
  } else {
    X.mid <- (X[seq(1, n - 1, by = 2), ] + X[seq(2, n, by = 2), ]) / 2
    mid.pairs <- matchNbp(X.mid)
    phi2.hat <-
      mean(Y.sum[mid.pairs[, 1]] * Y.sum[mid.pairs[, 2]])
  }

  Y.t <- Y[A == 1]
  Y.c <- Y[A == 0]
  var.sc <-
    mean((Y.t - mean(Y.t)) ^ 2) + mean((Y.c - mean(Y.c)) ^ 2) + (mean(Y.t) + mean(Y.c)) ^
    2 / 2 - phi2.hat / 2
  test.den <- sqrt(var.sc)
  test.stat <- test.num / test.den
  test <- abs(test.stat) > qnorm(1 - alpha.level / 2)
  list(test = test, test.num = test.num / sqrt(n), test.den = test.den / sqrt(n), n = n)
}
testSCNew <- cmpfun(testSCNew)

testSCNewRand <- function(data.obs, alpha.level = 0.05, n.perm) {
  # ------------------------
  # Randomization with correct studentization
  # -------------------------
  n <- data.obs$n
  X <- data.obs$X
  A <- data.obs$A
  Y <- data.obs$Y
  A.mat <- matrix(A, n, 2, byrow = T)
  Y.sum <- Y[A == 1] + Y[A == 0]
  test.num <- sqrt(n) * mean(Y[A == 1] - Y[A == 0])

  if (dim(X)[2] == 1) {
    phi2.hat <-
      mean(Y.sum[seq(1, n - 1, by = 2)] * Y.sum[seq(2, n, by = 2)])
  } else {
    X.mid <- (X[seq(1, n - 1, by = 2), ] + X[seq(2, n, by = 2), ]) / 2
    mid.pairs <- matchNbp(X.mid)
    phi2.hat <-
      mean(Y.sum[mid.pairs[, 1]] * Y.sum[mid.pairs[, 2]])
  }

  Y.t <- Y[A == 1]
  Y.c <- Y[A == 0]
  var.sc <-
    mean((Y.t - mean(Y.t)) ^ 2) + mean((Y.c - mean(Y.c)) ^ 2) + (mean(Y.t) + mean(Y.c)) ^
    2 / 2 - phi2.hat / 2
  test.den <- sqrt(var.sc)
  test.stat <- test.num / test.den
  perm.stat <- rep(NA, n.perm)

  for (i in 1:n.perm) {
    coin.flip <- sample(c(-1, 1), n, replace = T)
    A.perm.mat <- (A.mat - 1 / 2) * 2 * coin.flip
    A.perm.mat <- (A.perm.mat + 1) / 2
    A.perm <- matrix(t(A.perm.mat), 2 * n, 1)
    test.num <- sqrt(n) * mean(Y[A.perm == 1] - Y[A.perm == 0])
    Y.t <- Y[A.perm == 1]
    Y.c <- Y[A.perm == 0]
    var.sc <-
      mean((Y.t - mean(Y.t)) ^ 2) + mean((Y.c - mean(Y.c)) ^ 2) + (mean(Y.t) + mean(Y.c)) ^
      2 / 2 - phi2.hat / 2
    test.den <- sqrt(var.sc)
    perm.stat[i] <- test.num / test.den
  }
  
  test.cv <- quantile(perm.stat, 1 - alpha.level)
  test <- test.stat > test.cv
}
testSCNewRand <- cmpfun(testSCNewRand)